// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//---------------------------------------------------------------------------
//
//                     XMM Science Analysis System (SAS)
//                    (c) 1997-2010 European Space Agency
//
//---------------------------------------------------------------------------
//
//	DssBuilder.cc: --- implementation of DssBuilder methods
//
//	DESCRIPTION:
//		Implementation of all methods in DssBuilder class - no inlines
//
//	CHANGE HISTORY:
//		UL, 1999-11-16: 1.0.0: - first version
//
//	REFERENCES:
//		[1]: Documentation of selectlib package, U. Lammers
//
//	AUTHOR:
//		U. Lammers (UL), Tue Nov 16 09:56:55 MET 1999
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/dsslib/src/build/DssBuilder.cc,v 1.2 2009/05/18 11:32:48 aibarra Exp $
//
#include <sstream>

#include "Tracer.h"
#include "DssBuilder.h"
#include "DataFilter.h"
#include "errstr.h"

DssBuilder::DssBuilder() : _currentDss(0), _dssVector(0), _currentComp(0),
                           _region(0), _currentRegionComponent(0), _currentRegion(0), axisCount(0)
{
  PROC
  ENDPROC
} // DssBuilder::DssBuilder()


DataSubspace DssBuilder::dss()
{
	PROC

	initDss();

	RETURN *_currentDss;
} // DssBuilder::dss()


DataSubspace DssBuilder::dss(const std::string &expression, const Table *table)
{
	PROC

	//
	//	first expand the expression
	//

	std::string expexpr(expression);
	if (table)
		expexpr = DataFilter::expandSelectionExpression(expression, *table);

	//	initialize Dss

	initDss();

	//  DssParser::yydebug = 1;

	DssParser::lexinit(expexpr, table);	//	initialize the lexer ...
	if (DssParser::yyparse()) {		//	... and start parsing
	  selerrstr<< "Data subspace syntax error."
		   <<  selerror("DssSyntaxError", 1);
	  SELTHROW(InvalidExpression);
	}

	delete _currentDss;
	_currentDss = *(_dssVector.end()-1);
	_dssVector.pop_back();

	_currentDss->compressRegions();

	RETURN *_currentDss;
} // DssBuilder::dss()


DataSubspace DssBuilder::getDss(const std::string &expression, const Table *table /*= 0*/)
{
  DataSubspace dss, blankDss;

  selectErrorHandler.exceptions(true);   // enable exception

  try {
    dss = dssBuilder.dss(expression, table);

  } catch(DssBuilder::InvalidExpression except) {
    errstr<< except.text()
	  << " No data subspace information from this expression will be written to output."
	  << Sas::warning("DssException");
    dss = blankDss;
    dss.setError();

  } catch(DssBuilder::MiscException except) {
    errstr<< except.text()
	  << " No data subspace information from this expression will be written to output."
	  << Sas::warning("DssException");
    dss = blankDss;
    dss.setError();

  } catch(DssBuilder::ArithmeticException except) {
    errstr<< except.text()
	  << " No data subspace information from this expression will be written to output."
	  << Sas::warning("DssException");
    dss = blankDss;
    dss.setError();

  } catch (...) {
    errstr<<"This should not happen - raise an SPR. "
	  <<Sas::error("UnknownDssBuilderError");
  }

  //  selectErrorHandler.oldHandling();    // restore original error handling scheme

  clear();  // Free up the any allocated space and release any dss/dal objects.

  return dss;
}

//
//	PURPOSE:
//		Intialize the internally held DSS objects
//
void DssBuilder::initDss()
{
	PROC

	_currentDss = new DataSubspace;
	_dssVector.clear();

	_currentComp = 0;
	_region =  0;
	_currentRegionComponent =  0;
	_currentRegion = 0;

	ENDPROC
} // DssBuilder::initDss()


DataSubspace *DssBuilder::newDss() {
  _dssVector.push_back(_currentDss);
  _currentDss = new DataSubspace;

  newComponent();

  return _currentDss;
}


DataSubspace *DssBuilder::andDss() {
  DataSubspace * dss = *(_dssVector.end()-1);
  _dssVector.pop_back();

  **(_dssVector.end()-1) *= *dss;
  delete dss;
  return *(_dssVector.end()-1);
}

DataSubspace *DssBuilder::orDss() {
  DataSubspace * dss = *(_dssVector.end()-1);
  _dssVector.pop_back();

  for(unsigned icomp=0; icomp < dss->components(); icomp++)
    (*(_dssVector.end()-1))->addComponent(*dss->component(icomp));

  delete dss;
  return *(_dssVector.end()-1);
}

DataSubspace *DssBuilder::negateDss() {
  !(**(_dssVector.end()-1));
  return *(_dssVector.end()-1);
}


//
//	PURPOSE:
//		The parser is about to accept a new DSS component; add one to
//		the internal DSS object
//
DScomp *DssBuilder::newComponent()
{
	PROC

	_currentComp = _currentDss->addComponent();
	newRegion();

	RETURN _currentComp;
} // DssBuilder::newComponent()


GeoRegion *DssBuilder::newRegion() {
	initRegion();
	_region->push_back(Component());
	_currentRegionComponent = &(*_region)[_region->size()-1];
	_xAxisName = "";
	_yAxisName = "";

	return _region;
}


//
//	Purpose:
//		Return the current DSS component to the caller; this is needed by
//		the parser to add new filters
//
DScomp *DssBuilder::currentComponent()
{
	PROC

	if (!_currentComp)
		ALTRETURNV(newComponent());

	RETURN _currentComp;
} // DssBuilder::currentComponent()


//
//	DssBuilder::initRegion: initialize internal Region object
//
void DssBuilder::initRegion()
{
	PROC

	if (!_region)
		_region = new GeoRegion;
}


//
//	PURPOSE:
//		Return the current region component to the caller; this is needed
//		by the parser to add new region filters to the current region component
//
Component *DssBuilder::currentRegionComponent()
{
	PROC

	initRegion();
	if (!_currentRegionComponent) {
		_region->push_back(Component());
		_currentRegionComponent = &(*_region)[0];
	}

	RETURN _currentRegionComponent;
}


//
//	PURPOSE:
//		Return the most recently added region element (as an FShape *) to the
//		caller; this is needed by the parser in case is has detected that
//		an element is to be excluded (e.g. "!(X in circle(1,2,3))")
//
FShape *DssBuilder::currentRegion()
{
	PROC

	if (!_currentRegion) {
		selerrstr<<"No region has been registered yet."
				 <<selerror("NoCurrentRegion", 2);
		SELTHROW(MiscException);
	}

	RETURN _currentRegion;
} // DssBuilder::currentRegion()


//
//	PURPOSE:
//		Register name of 2-D region axes or compare axes names with
//		previously registered ones; this is to detect conflicts
//		with differnt axis names of different region in the same component
//
void DssBuilder::regionAxes(const std::string &a1, const std::string &a2)
{
	PROC

	_xAxisName = a1;
	_yAxisName = a2;
} // DssBuilder::regionAxes()


//
//	PURPOSE:
//		Add a region shape to the currently active region element
//
void DssBuilder::addRegion(FShape *region)
{
	PROC

	_currentRegion = region;
	currentRegionComponent()->push_back(region);

	ENDPROC
} // DssBuilder::addRegion()


//
//	PURPOSE:
//              Construct a reasonable name for a 2d axis, given the x and y component names
//
std::string DssBuilder::Axis2dName(const DSS::DSfilterComp &fcomp){
  if( fcomp[0]!="" && fcomp[1]!="" ){
    if( fcomp[0]=="X" && fcomp[1]=="Y" ) return "POS";
    if( fcomp[0]=="RA" && fcomp[1]=="DEC" ) return "EQPOS";
    if( fcomp[0]=="GLON" && fcomp[1]=="GLAT" ) return "GALPOS";
    if( fcomp[0]=="BETA_CORR" && fcomp[1]=="PI" ) return "BANANA";
    if( fcomp[0]=="BETA_CORR" && fcomp[1]=="XDSP_CORR" ) return "RGSPOS";

    if( fcomp[0].substr(fcomp[0].size()-1,1)=="X" &&
	fcomp[1].substr(fcomp[1].size()-1,1)=="Y" &&
	fcomp[0].substr(0,fcomp[0].size()-1) == fcomp[1].substr(0,fcomp[1].size()-1) )
      return fcomp[0].substr(0,fcomp[0].size()-1);
  }

  std::stringstream axisName;
  axisCount++;
  axisName << "AXIS2D" << axisCount;
  return axisName.str();
}


void DssBuilder::clear()
{
  //	be nice and free memory
  if (_region)
    for (unsigned n=0; n<_region->size(); ++n)
      for (unsigned m=0; m<(*_region)[n].size(); ++m)
	delete (*_region)[n][m];	//	delete region shapes that been allocated in parser

  delete _region; _region = 0;
  delete _currentDss; _currentDss = 0;
  _currentRegion = 0;
  _currentRegionComponent = 0;
  _currentComp = 0;
  for(unsigned i=0; i<_dssVector.size(); i++) delete _dssVector[i];

  _dssVector.clear();
}

//
//	PURPOSE:
//		DssBuilder destructor; frees all memory dynamically allocated
//		in the all of above methods or in the parser
DssBuilder::~DssBuilder()
{
  PROC
    clear();
  ENDPROC
} // DssBuilder::~DssBuilder()


//
//      PUPROSE:
//              Return a reference to a single global instance of a DssBuilder
//              object
//
DssBuilder &dssBuilderFunc()
{
        static DssBuilder *dssBuilderInstance = new DssBuilder;

        return *dssBuilderInstance;
} // dssBuilderFunc()


