// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//
//  Implementation of the DSgtiBLock class
//
//	AUTHOR:
//	  D. Hinshaw (DH)
//
//	  REFERENCES:
//	    [1] FITS Keyword Conventions in ASC Data Model files, SDS-7.2
// 		J. McDowell
//		http://hea-www.harvard.edu/~jcm/asc/docs/
//
//          [2] ASC Data Model Abstact Design, SDS-3.0
//              J. McDowell
//		http://hea-www.harvard.edu/~jcm/asc/docs/
//	    
//	    [3] Filtering Syntax in the ASC Data Model, SDS-6.0
//	        J. McDowell, A. Fruscione, A. Siemiginowska, B. Joye
//	        http://hea-www.harvard.edu/~jcm/asc/docs/

#include <cstdio>
#include <sstream>
#include <cstring>

#include <Dal.h>
#include <Region.h>
#include <DataSubspace.h>
#include <errstr.h>
#include <OGIPgti.h>


// Put this here temporarily until it gets put in the right place.
bool operator==(const Gti &gti1, const Gti &gti2){ return gti1.start()==gti2.start() && gti1.stop()==gti2.stop(); }


DSgtiBlock::DSgtiBlock(DSgtiBlock &dsblock) : DSblockFilter(dsblock) {}


DSgtiBlock::DSgtiBlock(DScomp *parent,
		       const std::string &name, const std::string &blockSpec,
		       const std::string &units /*= ""*/) :
  DSblockFilter(parent, GtiT, name, blockSpec, units){

    // Check that this block contains a valid gti table

    _openBlock();
    if( _block->type() == Block::TableT  &&
	_block->hasAttribute("HDUCLASS") &&
	_block->hasAttribute("HDUCLAS1") &&
	_block->attribute("HDUCLASS")->asString() == "OGIP" &&
	_block->attribute("HDUCLAS1")->asString() == "GTI" )
      return;
    else{
      Table * dsTable = dynamic_cast<Table *>(_block);
      if( dsTable->hasColumn("START") && dsTable->hasColumn("STOP") )
	return;
    }

    errstr << "Block type is not compatible with the gti filter type"
	   << Sas::error("BlockTypeGti");
}


DSgtiBlock::DSgtiBlock(DScomp *parent,
		       const Column *inheritedColumn, const std::string &blockSpec) :
  DSblockFilter(parent, GtiT, inheritedColumn, blockSpec){

    // Check that this block contains a valid gti table

    _openBlock();
    if( _block->type() == Block::TableT  &&
	_block->hasAttribute("HDUCLASS") &&
	_block->hasAttribute("HDUCLAS1") &&
	_block->attribute("HDUCLASS")->asString() == "OGIP" &&
	_block->attribute("HDUCLAS1")->asString() == "GTI" )
      return;
    else{
      Table * dsTable = dynamic_cast<Table *>(_block);
      if( dsTable->hasColumn("START") && dsTable->hasColumn("STOP") )
	return;
    }

    errstr << "Block type is not compatible with the gti filter type"
	   << Sas::error("BlockTypeGti");
}


std::string DSgtiBlock::selectExpression() {
  std::stringstream expression;
  expression << "gti(" << _parentName << ":" << _blockName << "," 
	     << _name << ")";
  return expression.str();
}


void DSgtiBlock::write(unsigned compNumber, unsigned filterNumber, 
		       Block *toBlock, bool dummy){
  // AI 15/11/2004
  // If the GTI are exactly the same only the first one is written to 
  // the event file.
  
  if(!dummy){
    // Check if there is already an identical GTI table in the output
    // data set.  If so, set the gti filter to point to that table.

    InOGIPgtiTable thisGtiTab(blockSpec());
    GtiList thisGtiList;
    thisGtiTab >> thisGtiList;

    DataSet * toSet = toBlock->parent();
    for(unsigned iblock=0; iblock<toSet->blocks(); iblock++)
      {
	Block * block = toSet->block(iblock);
	
	if( block->name() == _blockName     || 
	    !block->hasAttribute("HDUCLAS1") ||
	    block->attribute("HDUCLAS1")->asString() != "GTI" ) 
	  continue;
	
	InOGIPgtiTable gtiTab( dynamic_cast<Table *>(block) );
	GtiList gtiList;
	gtiTab >> gtiList;
	
	if(gtiList != thisGtiList) 
	  continue;
	
	// AI 15/11/2004
	// Do not change the GTI name altough the values were the same
	//_blockName = block->name();
	//_parentName = toSet->name();
	errstr << "Identical GTI in the table. "
	       << Sas::message(Msg::App, Msg::Verbose);
	
	break;
      }
  }
  
  DSblockFilter::write(compNumber, filterNumber, toBlock, dummy);
}


Range<double> DSgtiBlock::extent(const std::string &axis) {
  if( axis!=_name )
    errstr << "Attempt to extract extent information from the wrong filter"
	   << Sas::error("DSextentGTI");

  _openBlock();
  Table * table = dynamic_cast<Table *>(_block);

  ColumnData *lowerData = table->column("START")->data();
  ColumnData *upperData = table->column("STOP")->data();

  long unsigned nrows = table->rows();

  real64 *lowerVal = lowerData->real64Data();
  real64 *upperVal = upperData->real64Data();

  real64 min=lowerVal[0], max=upperVal[0];

  for(long unsigned irow=1; irow<nrows; irow++){
    if(lowerVal[irow]<min) min = lowerVal[irow];
    if(upperVal[irow]>max) max = upperVal[irow];
  }

  delete lowerVal;
  delete upperVal;

  Range<double> r(min,max);
  return r;
}

  
double DSgtiBlock::volume() {
  double vol=0;

  _openBlock();
  Table * table = dynamic_cast<Table *>(_block);

  ColumnData *lowerData = table->column("START")->data();
  ColumnData *upperData = table->column("STOP")->data();

  long unsigned nrows = table->rows();

  real64 *lowerVal = lowerData->real64Data();
  real64 *upperVal = upperData->real64Data();

  for(long unsigned irow=0; irow<nrows; irow++){
    vol += upperVal[irow]-lowerVal[irow]+1;
  }

  delete lowerVal;
  delete upperVal;

  return vol;  
}


void DSgtiBlock::operator!(){
  InOGIPgtiTable oldGtiTab(blockSpec());
  GtiList oldList, newList;
  oldGtiTab >> oldList;
  oldList.reduce();
  double start=0, stop;

  for(unsigned igti=0; igti<oldList.size(); igti++){
    stop = oldList[igti].start();
    if(stop!=0){
      Gti gti(start, stop);
      newList.push_back(gti);
    }
    start = oldList[igti].stop();
  }
  if( start!=Gti::ETERNITY ){
    stop = Gti::ETERNITY;
    Gti gti(start, stop);
    newList.push_back(gti);
  }

  DataSet * tempset = dataSetServer->open(tempnam(0,0), DataSet::Temp);
  OutOGIPgtiTable newGtiTab(tempset);
  newGtiTab << newList;

  DSgtiBlock newFilter(parent(), name(), tempset->name() + ":STDGTI", units());
  *this = newFilter;
}
